from __future__ import annotations

from marshmallow import fields

from zocalo.configuration import PluginSchema


class Slurm:
    class Schema(PluginSchema):
        url = fields.Str(required=True)
        user_token = fields.Str(required=False)
        user = fields.Str(required=False)
        api_version = fields.Str(required=True)

    @staticmethod
    def activate(configuration):
        return configuration



from __future__ import annotations

import os
import pathlib
from typing import Any

import requests

import zocalo.configuration

from . import models


class SlurmRestApi:
    def __init__(
        self,
        url: str,
        version: str = "v0.0.36",
        user_name: str | None = None,
        user_token: str | pathlib.Path | None = None,
    ):
        self.url = url
        self.version = version
        self.user_name = user_name
        if user_token and os.path.isfile(user_token):
            with open(user_token, "r") as f:
                self.user_token = f.read().strip()
        else:
            assert isinstance(user_token, str)
            self.user_token = user_token
        self.session = requests.Session()
        if self.user_name:
            self.session.headers["X-SLURM-USER-NAME"] = self.user_name
        if self.user_token:
            self.session.headers["X-SLURM-USER-TOKEN"] = self.user_token

    @classmethod
    def from_zocalo_configuration(cls, zc: zocalo.configuration.Configuration):
        return cls(
            url=zc.slurm["url"],
            version=zc.slurm["api_version"],
            user_name=zc.slurm.get("user"),
            user_token=zc.slurm.get("user_token"),
        )

    def get(
        self, endpoint: str, params: dict[str, Any] = None, timeout: float | None = None
    ) -> requests.Response:
        response = self.session.get(
            f"{self.url}/{endpoint}", params=params, timeout=timeout
        )
        response.raise_for_status()
        return response

    def put(
        self,
        endpoint: str,
        params: dict[str, Any] = None,
        json: dict[str, Any] = None,
        timeout: float | None = None,
    ) -> requests.Response:
        response = self.session.put(
            f"{self.url}/{endpoint}", params=params, json=json, timeout=timeout
        )
        response.raise_for_status()
        return response

    def post(
        self,
        endpoint: str,
        data: dict[str, Any] | None = None,
        json: dict[str, Any] | None = None,
        timeout: float | None = None,
    ) -> requests.Response:
        response = self.session.post(
            f"{self.url}/{endpoint}", data=data, json=json, timeout=timeout
        )
        response.raise_for_status()
        return response

    def delete(
        self, endpoint: str, params: dict[str, Any] = None, timeout: float | None = None
    ) -> requests.Response:
        response = self.session.delete(
            f"{self.url}/{endpoint}", params=params, timeout=timeout
        )
        response.raise_for_status()
        return response

    def get_jobs(self) -> models.JobsResponse:
        endpoint = f"slurm/{self.version}/jobs"
        response = self.get(endpoint)
        return models.JobsResponse(**response.json())

    def get_job_info(self, job_id: int) -> models.JobsResponse:
        endpoint = f"slurm/{self.version}/job/{job_id}"
        response = self.get(endpoint)
        return models.JobsResponse(**response.json())

    def submit_job(
        self, job_submission: models.JobSubmission
    ) -> models.JobSubmissionResponse:
        endpoint = f"slurm/{self.version}/job/submit"
        response = self.post(endpoint, json=job_submission.dict(exclude_defaults=True))
        return models.JobSubmissionResponse(**response.json())



from __future__ import annotations

import pytest

import zocalo.configuration

sample_configuration = """
version: 1

example:
  plugin: slurm
  url: http://example.com
  api_version: v0.0.36

example-with-username-and-token:
  plugin: slurm
  url: http://example.com:1234
  user: admin
  user_token: sometoken
  api_version: v0.0.36

environments:
  live:
    - example
  other:
    - example-with-username-and-token
"""


invalid_configuration = """
version: 1

invalid-example:
  plugin: slurm
  user: foo

environments:
  live:
    - invalid-example
"""


def test_without_username():
    zc = zocalo.configuration.from_string(sample_configuration)
    assert zc.slurm is None
    zc.activate_environment("live")
    assert isinstance(zc.slurm, dict)
    assert zc.slurm["url"] == "http://example.com"
    assert "user" not in zc.slurm
    assert zc.slurm["api_version"] == "v0.0.36"


def test_with_username():
    zc = zocalo.configuration.from_string(sample_configuration)
    assert zc.slurm is None
    zc.activate_environment("other")
    assert isinstance(zc.slurm, dict)
    assert zc.slurm["url"] == "http://example.com:1234"
    assert zc.slurm["user"] == "admin"
    assert zc.slurm["api_version"] == "v0.0.36"


def test_user_token_from_external_file(tmp_path):
    user_token = "abcdefg"
    user_token_file = tmp_path / "slurm-user-token"
    user_token_file.write_text(user_token)
    configuration = f"""
version: 1

example:
  plugin: slurm
  url: http://example.com
  api_version: v0.0.36
  user_token: {user_token_file}

environments:
  live:
    - example
"""
    zc = zocalo.configuration.from_string(configuration)
    assert zc.slurm is None
    zc.activate_environment("live")
    assert isinstance(zc.slurm, dict)
    assert zc.slurm["user_token"] == str(user_token_file)


def test_invalid_config():
    with pytest.raises(
        zocalo.ConfigurationError, match="Missing data for required field."
    ):
        zocalo.configuration.from_string(invalid_configuration)



from __future__ import annotations

import argparse
import json
import pathlib
import sys
import time

import workflows.transport

import zocalo.configuration


def run():
    zc = zocalo.configuration.from_file()
    zc.activate()
    dropdir = pathlib.Path(zc.storage["zocalo.go.fallback_location"])

    parser = argparse.ArgumentParser(
        usage="zocalo.pickup [options]", description="Processes zocalo.go backlog"
    )

    parser.add_argument("-?", action="help", help=argparse.SUPPRESS)
    parser.add_argument(
        "-d",
        "--delay",
        dest="delay",
        action="store",
        type=int,
        default=2,
        help="Number of seconds to wait between message dispatches",
    )
    parser.add_argument(
        "-w",
        "--wait",
        dest="wait",
        action="store",
        type=int,
        default=60,
        help="Number of seconds to wait initially",
    )
    parser.add_argument(
        "-v",
        "--verbose",
        dest="verbose",
        action="store_true",
        default=False,
        help="Show raw message before sending",
    )
    zc.add_command_line_options(parser)
    workflows.transport.add_command_line_options(parser, transport_argument=True)
    args = parser.parse_args()

    try:
        files = list(dropdir.iterdir())
    except OSError:
        sys.exit("This program is only available to privileged users")

    print(f"Found {len(files)} files")
    if not files:
        sys.exit()

    if args.wait:
        print(f"Waiting {args.wait} seconds")
        time.sleep(args.wait)

    print(f"Connecting to {args.transport}...")
    transport = workflows.transport.lookup(args.transport)()
    transport.connect()

    file_info = {f: {} for f in files}

    for f, finfo in file_info.items():
        with f.open() as fh:
            data = json.load(fh)
            finfo["message"] = data["message"]
            finfo["headers"] = data["headers"]
        finfo["originating-host"] = finfo["headers"].get("zocalo.go.host")
        finfo["recipes"] = ",".join(finfo["message"].get("recipes", []))
        finfo["mtime"] = f.stat().st_mtime

    count = 0
    file_count = len(file_info)
    for f, finfo in dict(
        sorted(file_info.items(), key=lambda item: item[1]["mtime"])
    ).items():
        print(
            f"Sending {f} from host {file_info[f]['originating-host']}"
            f" with recipes {file_info[f]['recipes']}"
        )
        assert f.exists()
        transport.send(
            "processing_recipe",
            file_info[f]["message"],
            headers=file_info[f]["headers"],
        )
        f.unlink()
        count = count + 1
        print(f"Done ({count} of {file_count})")
        try:
            time.sleep(args.delay)
        except KeyboardInterrupt:
            print("CTRL+C - stopping")
            time.sleep(0.5)
            sys.exit(1)

    transport.disconnect()



from __future__ import annotations

import copy
import errno
import json
import os
import re
import time
import timeit
import uuid

import pkg_resources
import workflows.recipe
from workflows.services.common_service import CommonService


class Dispatcher(CommonService):
    """
    Single point of contact service that takes in job meta-information
    (say, a data collection ID), a processing recipe, a list of recipes,
    or pointers to recipes stored elsewhere, and mangles these into something
    that can be processed by downstream services.
    """

    # Human readable service name
    _service_name = "Dispatcher"

    # Logger name
    _logger_name = "zocalo.service.dispatcher"

    def filter_load_recipes_from_files(self, message, parameters):
        """Load named recipes from central location and merge them into the recipe object"""
        for recipefile in message.get("recipes", []):
            try:
                with open(
                    os.path.join(self.recipe_basepath, recipefile + ".json"), "r"
                ) as rcp:
                    named_recipe = workflows.recipe.Recipe(recipe=rcp.read())
            except ValueError:
                raise ValueError(f"Error reading recipe {recipefile}")
            except IOError as e:
                if e.errno == errno.ENOENT:
                    raise ValueError(
                        f"Message references non-existing recipe {recipefile}. Recipe path is {self.recipe_basepath}",
                    )
                raise
            try:
                named_recipe.validate()
            except workflows.Error as e:
                raise ValueError(f"Named recipe {recipefile} failed validation. {e}")
            message["recipe"] = message["recipe"].merge(named_recipe)
        return message, parameters

    def filter_load_custom_recipe(self, message, parameters):
        """Load a custom recipe from a message and merge them into the recipe object"""
        if message.get("custom_recipe"):
            try:
                custom_recipe = workflows.recipe.Recipe(
                    recipe=json.dumps(message["custom_recipe"])
                )
                self.log.info(
                    "Received message containing a custom recipe: %s",
                    message["custom_recipe"],
                )
            except Exception as e:
                raise ValueError(
                    f"Error reading custom recipe {message['custom_recipe']}"
                ) from e
            try:
                custom_recipe.validate()
            except workflows.Error as e:
                raise ValueError(
                    f"Custom recipe {custom_recipe} failed validation. {e}"
                )
            message["recipe"] = message["recipe"].merge(custom_recipe)
        return message, parameters

    def filter_apply_parameters(self, message, parameters):
        """Fill in any placeholders in the recipe of the form {name} using the
        parameters data structure"""
        message["recipe"].apply_parameters(parameters)
        return message, parameters

    def initializing(self):
        """Subscribe to the processing_recipe queue. Received messages must be acknowledged."""
        self.log.info("Dispatcher starting")
        self.recipe_basepath = self._environment["config"].storage.get(
            "zocalo.recipe_directory"
        )
        # Store a copy of all dispatch messages in this location
        self._logbook = self._environment["config"].storage.get(
            "zocalo.dispatcher.logbook_location"
        )
        if self._logbook:
            try:
                os.makedirs(self._logbook, 0o775)
            except OSError:
                pass  # Ignore if exists
            if os.access(self._logbook, os.R_OK | os.W_OK | os.X_OK):
                self.log.debug(f"Using logbook location {self._logbook}")
            else:
                self.log.error(f"Logbook disabled: Can not write to {self._logbook}")
                self._logbook = None
        else:
            self.log.info(
                "Logbook disabled: zocalo.dispatcher.logbook_location not defined"
            )
            self._logbook = None

        self.message_filters = {
            **{
                f.name: f.load()
                for f in pkg_resources.iter_entry_points(
                    "zocalo.services.dispatcher.filters"
                )
            },
            "load_custom_recipe": self.filter_load_custom_recipe,
            "load_recipes_from_files": self.filter_load_recipes_from_files,
            "apply_parameters": self.filter_apply_parameters,
        }

        self.ready_for_processing = {
            f.name: f.load()
            for f in pkg_resources.iter_entry_points(
                "zocalo.services.dispatcher.ready_for_processing"
            )
        }

        workflows.recipe.wrap_subscribe(
            self._transport,
            "processing_recipe",
            self.process,
            acknowledgement=True,
            log_extender=self.extend_log,
            allow_non_recipe_messages=True,
        )

    def record_to_logbook(self, guid, header, original_message, message, recipewrap):
        basepath = os.path.join(self._logbook, time.strftime("%Y-%m"))
        clean_guid = re.sub(r"[^a-z0-9A-Z\-]+", "", guid, re.UNICODE)
        if not clean_guid or len(clean_guid) < 3:
            self.log.warning(
                "Message with non-conforming guid %s not written to logbook", guid
            )
            return
        try:
            os.makedirs(os.path.join(basepath, clean_guid[:2]))
        except OSError:
            pass  # Ignore if exists

        def neat_json_to_file(obj, fh, **kwargs):
            def _fix(item):
                if isinstance(item, list):
                    return [_fix(i) for i in item]
                if isinstance(item, dict):
                    return {str(key): _fix(value) for key, value in item.items()}
                return item

            return json.dump(
                _fix(obj),
                fh,
                sort_keys=True,
                skipkeys=True,
                default=str,
                indent=2,
                separators=(",", ": "),
                **kwargs,
            )

        try:
            log_entry = os.path.join(basepath, clean_guid[:2], clean_guid[2:])
            with open(log_entry, "w") as fh:
                fh.write("Incoming message header:\n")
                neat_json_to_file(header, fh)
                fh.write("\n\nIncoming message body:\n")
                neat_json_to_file(original_message, fh)
                fh.write("\n\nParsed message body:\n")
                neat_json_to_file(message, fh)
                fh.write("\n\nRecipe object:\n")
                neat_json_to_file(
                    recipewrap.recipe.recipe,
                    fh,
                )
                fh.write("\n")
            self.log.debug("Message saved in logbook at %s", log_entry)
        except Exception:
            self.log.warning("Could not write message to logbook", exc_info=True)

    def process(self, rw, header, message):
        """Process an incoming processing request."""
        # Time execution
        start_time = timeit.default_timer()

        # Load processing parameters
        parameters = message.get("parameters", {})
        if not isinstance(parameters, dict):
            # malformed message
            self.log.error(
                "Dispatcher rejected malformed message: parameters not given as dictionary"
            )
            self._transport.nack(header)
            return

        # Unless 'guid' is already defined then generate a unique recipe IDs for
        # this request, which is attached to all downstream log records and can
        # be used to determine unique file paths.
        recipe_id = parameters.get("guid") or str(uuid.uuid4())
        parameters["guid"] = recipe_id

        if rw:
            # If we received a recipe wrapper then we already have a recipe_ID
            # attached to logs. Make a note of the downstream recipe ID so that
            # we can track execution beyond recipe boundaries.
            self.log.info(
                "Processing request with new recipe ID %s:\n%s", recipe_id, str(message)
            )

        # If we are fully logging requests then make a copy of the original message
        if self._logbook:
            original_message = copy.deepcopy(message)

        # From here on add the global ID to all log messages
        with self.extend_log("recipe_ID", recipe_id):
            self.log.debug("Received processing request:\n" + str(message))
            self.log.debug("Received processing parameters:\n" + str(parameters))

            # Step 1: Check that parsing the message can proceed
            for name, ready_for_processing in self.ready_for_processing.items():
                if not ready_for_processing(message, parameters):
                    # Message not yet cleared for processing
                    if "dispatcher_expiration" not in parameters:
                        parameters["dispatcher_expiration"] = time.time() + int(
                            parameters.get("dispatcher_timeout", 120)
                        )
                    if parameters["dispatcher_expiration"] > time.time():
                        # Wait for 2 seconds
                        txn = self._transport.transaction_begin(
                            subscription_id=header["subscription"]
                        )
                        self._transport.ack(header, transaction=txn)
                        self._transport.send(
                            "processing_recipe", message, transaction=txn, delay=2
                        )
                        self.log.info("Message not yet ready for processing")
                        self._transport.transaction_commit(txn)
                        return
                    elif parameters.get("dispatcher_error_queue"):
                        # Drop message into error queue
                        txn = self._transport.transaction_begin(
                            subscription_id=header["subscription"]
                        )
                        self._transport.ack(header, transaction=txn)
                        self._transport.send(
                            parameters["dispatcher_error_queue"],
                            message,
                            transaction=txn,
                        )
                        self.log.info(
                            "Message rejected to specified error queue as still not ready for processing"
                        )
                        self._transport.transaction_commit(txn)
                        return
                    else:
                        # Unhandled error, send message to DLQ
                        self.log.error(
                            "Message rejected as still not ready for processing",
                        )
                        self._transport.nack(header)
                        return

            filtered_message = copy.deepcopy(message)
            filtered_parameters = copy.deepcopy(parameters)

            # Create empty recipe
            filtered_message["recipe"] = workflows.recipe.Recipe()

            # Apply all specified filters in order to message and parameters
            for name, f in self.message_filters.items():
                try:
                    filtered_message, filtered_parameters = f(
                        filtered_message, filtered_parameters
                    )
                except Exception as e:
                    self.log.error(
                        "Rejected message due to filter (%s) error: %s",
                        name,
                        str(e),
                        exc_info=True,
                    )
                    self._transport.nack(header)
                    return

            self.log.debug("Mangled processing request:\n" + str(filtered_message))
            self.log.debug(
                "Mangled processing parameters:\n" + str(filtered_parameters)
            )

            # Conditionally acknowledge receipt of the message
            txn = self._transport.transaction_begin(
                subscription_id=header["subscription"]
            )
            self._transport.ack(header, transaction=txn)

            rw = workflows.recipe.RecipeWrapper(
                recipe=filtered_message["recipe"], transport=self._transport
            )
            rw.environment = {
                "ID": recipe_id
            }  # FIXME: This should go into the constructor, but workflows can't do that yet
            rw.start(transaction=txn)

            # Write information to logbook if applicable
            if self._logbook:
                self.record_to_logbook(
                    recipe_id, header, original_message, filtered_message, rw
                )

            # Commit transaction
            self._transport.transaction_commit(txn)
            self.log.info(
                "Processed incoming message in %.4f seconds",
                timeit.default_timer() - start_time,
            )



from __future__ import annotations

from enum import Enum
from typing import Any, Dict, List, Optional

from pydantic import BaseModel, Field


class Exclusive(Enum):
    user = "user"
    mcs = "mcs"
    true = "true"
    false = "false"


class GresFlags(Enum):
    disable_binding = "disable-binding"
    enforce_binding = "enforce-binding"


class OpenMode(Enum):
    append = "append"
    truncate = "truncate"


class Error(BaseModel):
    error: Optional[str] = Field(None, description="error message")
    errno: Optional[int] = Field(None, description="error number")


class JobProperties(BaseModel):
    class Config:
        use_enum_values = True

    account: Optional[str] = Field(
        None, description="Charge resources used by this job to specified account."
    )
    account_gather_freqency: Optional[str] = Field(
        None, description="Define the job accounting and profiling sampling intervals."
    )
    argv: Optional[List[str]] = Field(None, description="Arguments to the script.")
    array: Optional[str] = Field(
        None,
        description="Submit a job array, multiple jobs to be executed with identical parameters. The indexes specification identifies what array index values should be used.",
    )
    batch_features: Optional[str] = Field(
        None, description="features required for batch script's node"
    )
    begin_time: Optional[str] = Field(
        None,
        description="Submit the batch script to the Slurm controller immediately, like normal, but tell the controller to defer the allocation of the job until the specified time.",
    )
    burst_buffer: Optional[str] = Field(None, description="Burst buffer specification.")
    cluster_constraints: Optional[str] = Field(
        None,
        description="Specifies features that a federated cluster must have to have a sibling job submitted to it.",
    )
    comment: Optional[str] = Field(None, description="An arbitrary comment.")
    constraints: Optional[str] = Field(
        None, description="node features required by job."
    )
    core_specification: Optional[int] = Field(
        None,
        description="Count of specialized threads per node reserved by the job for system operations and not used by the application.",
    )
    cores_per_socket: Optional[int] = Field(
        None,
        description="Restrict node selection to nodes with at least the specified number of cores per socket.",
    )
    cpu_binding: Optional[str] = Field(None, description="Cpu binding")
    cpu_binding_hint: Optional[str] = Field(None, description="Cpu binding hint")
    cpu_frequency: Optional[str] = Field(
        None,
        description="Request that job steps initiated by srun commands inside this sbatch script be run at some requested frequency if possible, on the CPUs selected for the step on the compute node(s).",
    )
    cpus_per_gpu: Optional[str] = Field(
        None, description="Number of CPUs requested per allocated GPU."
    )
    cpus_per_task: Optional[int] = Field(
        None,
        description="Advise the Slurm controller that ensuing job steps will require ncpus number of processors per task.",
    )
    current_working_directory: Optional[str] = Field(
        None,
        description="Instruct Slurm to connect the batch script's standard output directly to the file name.",
    )
    deadline: Optional[str] = Field(
        None,
        description="Remove the job if no ending is possible before this deadline (start > (deadline - time[-min])).",
    )
    delay_boot: Optional[int] = Field(
        None,
        description="Do not reboot nodes in order to satisfied this job's feature specification if the job has been eligible to run for less than this time period.",
    )
    dependency: Optional[str] = Field(
        None,
        description="Defer the start of this job until the specified dependencies have been satisfied completed.",
    )
    distribution: Optional[str] = Field(
        None, description="Specify alternate distribution methods for remote processes."
    )
    environment: Dict[str, Any] = Field(
        ..., description="Dictionary of environment entries."
    )
    exclusive: Optional[Exclusive] = Field(
        None,
        description='The job allocation can share nodes just other users with the "user" option or with the "mcs" option).',
    )
    get_user_environment: Optional[bool] = Field(
        None, description="Load new login environment for user on job node."
    )
    gres: Optional[str] = Field(
        None,
        description="Specifies a comma delimited list of generic consumable resources.",
    )
    gres_flags: Optional[GresFlags] = Field(
        None, description="Specify generic resource task binding options."
    )
    gpu_binding: Optional[str] = Field(
        None, description="Requested binding of tasks to GPU."
    )
    gpu_frequency: Optional[str] = Field(None, description="Requested GPU frequency.")
    gpus: Optional[str] = Field(None, description="GPUs per job.")
    gpus_per_node: Optional[str] = Field(None, description="GPUs per node.")
    gpus_per_socket: Optional[str] = Field(None, description="GPUs per socket.")
    gpus_per_task: Optional[str] = Field(None, description="GPUs per task.")
    hold: Optional[bool] = Field(
        None,
        description="Specify the job is to be submitted in a held state (priority of zero).",
    )
    kill_on_invalid_dependency: Optional[bool] = Field(
        None,
        description="If a job has an invalid dependency, then Slurm is to terminate it.",
    )
    licenses: Optional[str] = Field(
        None,
        description="Specification of licenses (or other resources available on all nodes of the cluster) which must be allocated to this job.",
    )
    mail_type: Optional[str] = Field(
        None, description="Notify user by email when certain event types occur."
    )
    mail_user: Optional[str] = Field(
        None,
        description="User to receive email notification of state changes as defined by mail_type.",
    )
    mcs_label: Optional[str] = Field(
        None, description="This parameter is a group among the groups of the user."
    )
    memory_binding: Optional[str] = Field(None, description="Bind tasks to memory.")
    memory_per_cpu: Optional[int] = Field(
        None, description="Minimum real memory per cpu (MB)."
    )
    memory_per_gpu: Optional[int] = Field(
        None, description="Minimum memory required per allocated GPU."
    )
    memory_per_node: Optional[int] = Field(
        None, description="Minimum real memory per node (MB)."
    )
    minimum_cpus_per_node: Optional[int] = Field(
        None, description="Minimum number of CPUs per node."
    )
    minimum_nodes: Optional[bool] = Field(
        None,
        description="If a range of node counts is given, prefer the smaller count.",
    )
    name: Optional[str] = Field(
        None, description="Specify a name for the job allocation."
    )
    nice: Optional[str] = Field(
        None,
        description="Run the job with an adjusted scheduling priority within Slurm.",
    )
    no_kill: Optional[bool] = Field(
        None,
        description="Do not automatically terminate a job if one of the nodes it has been allocated fails.",
    )
    nodes: Optional[List[int]] = Field(
        None,
        description="Request that a minimum of minnodes nodes and a maximum node count.",
        max_items=2,
        # min_items=1, XXX
        min_items=2,
    )
    open_mode: Optional[OpenMode] = Field(
        "append",
        description="Open the output and error files using append or truncate mode as specified.",
    )
    partition: Optional[str] = Field(
        None, description="Request a specific partition for the resource allocation."
    )
    priority: Optional[str] = Field(
        None, description="Request a specific job priority."
    )
    qos: Optional[str] = Field(
        None, description="Request a quality of service for the job."
    )
    requeue: Optional[bool] = Field(
        None,
        description="Specifies that the batch job should eligible to being requeue.",
    )
    reservation: Optional[str] = Field(
        None, description="Allocate resources for the job from the named reservation."
    )
    signal: Optional[str] = Field(
        None,
        description="When a job is within sig_time seconds of its end time, send it the signal sig_num.",
        regex="[B:]<sig_num>[@<sig_time>]",
    )
    sockets_per_node: Optional[int] = Field(
        None,
        description="Restrict node selection to nodes with at least the specified number of sockets.",
    )
    spread_job: Optional[bool] = Field(
        None,
        description="Spread the job allocation over as many nodes as possible and attempt to evenly distribute tasks across the allocated nodes.",
    )
    standard_error: Optional[str] = Field(
        None,
        description="Instruct Slurm to connect the batch script's standard error directly to the file name.",
    )
    standard_in: Optional[str] = Field(
        None,
        description="Instruct Slurm to connect the batch script's standard input directly to the file name specified.",
    )
    standard_out: Optional[str] = Field(
        None,
        description="Instruct Slurm to connect the batch script's standard output directly to the file name.",
    )
    tasks: Optional[int] = Field(
        None,
        description="Advises the Slurm controller that job steps run within the allocation will launch a maximum of number tasks and to provide for sufficient resources.",
    )
    tasks_per_core: Optional[int] = Field(
        None, description="Request the maximum ntasks be invoked on each core."
    )
    tasks_per_node: Optional[int] = Field(
        None, description="Request the maximum ntasks be invoked on each node."
    )
    tasks_per_socket: Optional[int] = Field(
        None, description="Request the maximum ntasks be invoked on each socket."
    )
    thread_specification: Optional[int] = Field(
        None,
        description="Count of specialized threads per node reserved by the job for system operations and not used by the application.",
    )
    threads_per_core: Optional[int] = Field(
        None,
        description="Restrict node selection to nodes with at least the specified number of threads per core.",
    )
    time_limit: Optional[int] = Field(None, description="Step time limit.")
    time_minimum: Optional[int] = Field(
        None, description="Minimum run time in minutes."
    )
    wait_all_nodes: Optional[bool] = Field(
        None, description="Do not begin execution until all nodes are ready for use."
    )
    wckey: Optional[str] = Field(None, description="Specify wckey to be used with job.")


class JobSubmission(BaseModel):
    script: str = Field(
        ..., description="Executable script (full contents) to run in batch step"
    )
    job: Optional[JobProperties] = Field(
        None, description="Properties of an array job or non-HetJob"
    )
    jobs: Optional[List[JobProperties]] = Field(
        None, description="Properties of an HetJob"
    )


class JobSubmissionResponse(BaseModel):
    errors: Optional[List[Error]] = Field(None, description="slurm errors")
    job_id: Optional[int] = Field(None, description="new job ID")
    step_id: Optional[str] = Field(None, description="new job step ID")
    job_submit_user_msg: Optional[str] = Field(
        None, description="Message to user from job_submit plugin"
    )


class NodeAllocation(BaseModel):
    memory: Optional[int] = Field(None, description="amount of assigned job memory")
    cpus: Optional[int] = Field(None, description="amount of assigned job CPUs")
    sockets: Optional[Dict[str, Any]] = Field(
        None, description="assignment status of each socket by socket id"
    )
    cores: Optional[Dict[str, Any]] = Field(
        None, description="assignment status of each core by core id"
    )


class JobResources(BaseModel):
    nodes: Optional[str] = Field(None, description="list of assigned job nodes")
    allocated_cpus: Optional[int] = Field(
        None, description="number of assigned job cpus"
    )
    allocated_hosts: Optional[int] = Field(
        None, description="number of assigned job hosts"
    )
    allocated_nodes: Optional[Dict[str, NodeAllocation]] = Field(
        None, description="node allocations"
    )


class JobResponseProperties(BaseModel):
    account: Optional[str] = Field(
        None, description="Charge resources used by this job to specified account"
    )
    accrue_time: Optional[str] = Field(
        None, description="time job is eligible for running"
    )
    admin_comment: Optional[str] = Field(
        None, description="administrator's arbitrary comment"
    )
    array_job_id: Optional[str] = Field(
        None, description="job_id of a job array or 0 if N/A"
    )
    array_task_id: Optional[str] = Field(None, description="task_id of a job array")
    array_max_tasks: Optional[str] = Field(
        None, description="Maximum number of running array tasks"
    )
    array_task_string: Optional[str] = Field(
        None, description="string expression of task IDs in this record"
    )
    association_id: Optional[str] = Field(None, description="association id for job")
    batch_features: Optional[str] = Field(
        None, description="features required for batch script's node"
    )
    batch_flag: Optional[bool] = Field(
        None, description="if batch: queued job with script"
    )
    batch_host: Optional[str] = Field(
        None, description="name of host running batch script"
    )
    flags: Optional[List[str]] = Field(None, description="Job flags")
    burst_buffer: Optional[str] = Field(None, description="burst buffer specifications")
    burst_buffer_state: Optional[str] = Field(
        None, description="burst buffer state info"
    )
    cluster: Optional[str] = Field(
        None, description="name of cluster that the job is on"
    )
    cluster_features: Optional[str] = Field(
        None, description="comma separated list of required cluster features"
    )
    command: Optional[str] = Field(None, description="command to be executed")
    comment: Optional[str] = Field(None, description="arbitrary comment")
    contiguous: Optional[bool] = Field(
        None, description="job requires contiguous nodes"
    )
    core_spec: Optional[str] = Field(None, description="specialized core count")
    thread_spec: Optional[str] = Field(None, description="specialized thread count")
    cores_per_socket: Optional[str] = Field(
        None, description="cores per socket required by job"
    )
    billable_tres: Optional[str] = Field(None, description="billable TRES")
    cpus_per_task: Optional[str] = Field(
        None, description="number of processors required for each task"
    )
    cpu_frequency_minimum: Optional[str] = Field(
        None, description="Minimum cpu frequency"
    )
    cpu_frequency_maximum: Optional[str] = Field(
        None, description="Maximum cpu frequency"
    )
    cpu_frequency_governor: Optional[str] = Field(
        None, description="cpu frequency governor"
    )
    cpus_per_tres: Optional[str] = Field(
        None, description="semicolon delimited list of TRES=# values"
    )
    deadline: Optional[str] = Field(None, description="job start deadline ")
    delay_boot: Optional[str] = Field(None, description="command to be executed")
    dependency: Optional[str] = Field(
        None, description="synchronize job execution with other jobs"
    )
    derived_exit_code: Optional[str] = Field(
        None, description="highest exit code of all job steps"
    )
    eligible_time: Optional[str] = Field(
        None, description="time job is eligible for running"
    )
    end_time: Optional[str] = Field(
        None, description="time of termination, actual or expected"
    )
    excluded_nodes: Optional[str] = Field(
        None, description="comma separated list of excluded nodes"
    )
    exit_code: Optional[int] = Field(None, description="exit code for job")
    features: Optional[str] = Field(
        None, description="comma separated list of required features"
    )
    federation_origin: Optional[str] = Field(None, description="Origin cluster's name")
    federation_siblings_active: Optional[str] = Field(
        None, description="string of active sibling names"
    )
    federation_siblings_viable: Optional[str] = Field(
        None, description="string of viable sibling names"
    )
    gres_detail: Optional[List[str]] = Field(None, description="Job flags")
    group_id: Optional[str] = Field(None, description="group job submitted as")
    job_id: Optional[str] = Field(None, description="job ID")
    job_resources: Optional[JobResources] = None
    job_state: Optional[str] = Field(None, description="state of the job")
    last_sched_evaluation: Optional[str] = Field(
        None, description="last time job was evaluated for scheduling"
    )
    licenses: Optional[str] = Field(None, description="licenses required by the job")
    max_cpus: Optional[str] = Field(
        None, description="maximum number of cpus usable by job"
    )
    max_nodes: Optional[str] = Field(
        None, description="maximum number of nodes usable by job"
    )
    mcs_label: Optional[str] = Field(None, description="mcs_label if mcs plugin in use")
    memory_per_tres: Optional[str] = Field(
        None, description="semicolon delimited list of TRES=# values"
    )
    name: Optional[str] = Field(None, description="name of the job")
    nodes: Optional[str] = Field(None, description="list of nodes allocated to job")
    nice: Optional[str] = Field(None, description="requested priority change")
    tasks_per_core: Optional[str] = Field(
        None, description="number of tasks to invoke on each core"
    )
    tasks_per_socket: Optional[str] = Field(
        None, description="number of tasks to invoke on each socket"
    )
    tasks_per_board: Optional[str] = Field(
        None, description="number of tasks to invoke on each board"
    )
    cpus: Optional[str] = Field(
        None, description="minimum number of cpus required by job"
    )
    node_count: Optional[str] = Field(
        None, description="minimum number of nodes required by job"
    )
    tasks: Optional[str] = Field(None, description="requested task count")
    het_job_id: Optional[str] = Field(None, description="job ID of hetjob leader")
    het_job_id_set: Optional[str] = Field(
        None, description="job IDs for all components"
    )
    het_job_offset: Optional[str] = Field(
        None, description="HetJob component offset from leader"
    )
    partition: Optional[str] = Field(None, description="name of assigned partition")
    memory_per_node: Optional[str] = Field(
        None, description="minimum real memory per node"
    )
    memory_per_cpu: Optional[str] = Field(
        None, description="minimum real memory per cpu"
    )
    minimum_cpus_per_node: Optional[str] = Field(
        None, description="minimum # CPUs per node"
    )
    minimum_tmp_disk_per_node: Optional[str] = Field(
        None, description="minimum tmp disk per node"
    )
    preempt_time: Optional[str] = Field(None, description="preemption signal time")
    pre_sus_time: Optional[str] = Field(
        None, description="time job ran prior to last suspend"
    )
    priority: Optional[str] = Field(None, description="relative priority of the job")
    profile: Optional[List[str]] = Field(None, description="Job profiling requested")
    qos: Optional[str] = Field(None, description="Quality of Service")
    reboot: Optional[bool] = Field(
        None, description="node reboot requested before start"
    )
    required_nodes: Optional[str] = Field(
        None, description="comma separated list of required nodes"
    )
    requeue: Optional[bool] = Field(
        None, description="enable or disable job requeue option"
    )
    resize_time: Optional[str] = Field(None, description="time of latest size change")
    restart_cnt: Optional[str] = Field(None, description="count of job restarts")
    resv_name: Optional[str] = Field(None, description="reservation name")
    shared: Optional[str] = Field(
        None, description="type and if job can share nodes with other jobs"
    )
    show_flags: Optional[List[str]] = Field(None, description="details requested")
    sockets_per_board: Optional[str] = Field(
        None, description="sockets per board required by job"
    )
    sockets_per_node: Optional[str] = Field(
        None, description="sockets per node required by job"
    )
    start_time: Optional[str] = Field(
        None, description="time execution begins, actual or expected"
    )
    state_description: Optional[str] = Field(
        None, description="optional details for state_reason"
    )
    state_reason: Optional[str] = Field(
        None, description="reason job still pending or failed"
    )
    standard_error: Optional[str] = Field(
        None, description="pathname of job's stderr file"
    )
    standard_input: Optional[str] = Field(
        None, description="pathname of job's stdin file"
    )
    standard_output: Optional[str] = Field(
        None, description="pathname of job's stdout file"
    )
    submit_time: Optional[str] = Field(None, description="time of job submission")
    suspend_time: Optional[str] = Field(
        None, description="time job last suspended or resumed"
    )
    system_comment: Optional[str] = Field(
        None, description="slurmctld's arbitrary comment"
    )
    time_limit: Optional[str] = Field(None, description="maximum run time in minutes")
    time_minimum: Optional[str] = Field(None, description="minimum run time in minutes")
    threads_per_core: Optional[str] = Field(
        None, description="threads per core required by job"
    )
    tres_bind: Optional[str] = Field(
        None, description="Task to TRES binding directives"
    )
    tres_freq: Optional[str] = Field(None, description="TRES frequency directives")
    tres_per_job: Optional[str] = Field(
        None, description="semicolon delimited list of TRES=# values"
    )
    tres_per_node: Optional[str] = Field(
        None, description="semicolon delimited list of TRES=# values"
    )
    tres_per_socket: Optional[str] = Field(
        None, description="semicolon delimited list of TRES=# values"
    )
    tres_per_task: Optional[str] = Field(
        None, description="semicolon delimited list of TRES=# values"
    )
    tres_req_str: Optional[str] = Field(None, description="tres reqeusted in the job")
    tres_alloc_str: Optional[str] = Field(None, description="tres used in the job")
    user_id: Optional[str] = Field(None, description="user id the job runs as")
    user_name: Optional[str] = Field(None, description="user the job runs as")
    wckey: Optional[str] = Field(None, description="wckey for job")
    current_working_directory: Optional[str] = Field(
        None, description="pathname of working directory"
    )


class JobsResponse(BaseModel):
    errors: Optional[List[Error]] = Field(None, description="slurm errors")
    jobs: Optional[List[JobResponseProperties]] = Field(
        None, description="job descriptions"
    )



from __future__ import annotations

import json
import sys
import time
import uuid
from unittest import mock

import pytest
import workflows.transport
from workflows.transport.common_transport import CommonTransport

import zocalo.configuration
from zocalo.cli import pickup


@pytest.fixture
def mock_zocalo_configuration(tmp_path):
    mock_zc = mock.MagicMock(zocalo.configuration.Configuration)
    mock_zc.storage = {
        "zocalo.go.fallback_location": str(tmp_path),
    }
    return mock_zc


def test_pickup_empty_filelist_raises_system_exit(mocker, mock_zocalo_configuration):
    mocker.patch.object(
        zocalo.configuration, "from_file", return_value=mock_zocalo_configuration
    )
    with mock.patch.object(sys, "argv", ["prog"]), pytest.raises(SystemExit) as e:
        pickup.run()
        assert e.code == 0


def test_pickup_sends_to_processing_recipe(mocker, mock_zocalo_configuration, tmp_path):
    mocked_transport = mocker.MagicMock(CommonTransport)
    mocker.patch.object(workflows.transport, "lookup", return_value=mocked_transport)
    mocker.patch.object(
        zocalo.configuration, "from_file", return_value=mock_zocalo_configuration
    )
    for i in range(10):
        msg = {
            "headers": {"zocalo.go.user": "foobar", "zocalo.go.host": "example.com"},
            "message": {
                "recipes": [f"thing{i}"],
                "parameters": {"foo": i},
            },
        }
        (tmp_path / str(uuid.uuid4())).write_text(json.dumps(msg))
        time.sleep(0.1)
    with mock.patch.object(sys, "argv", ["prog", "--wait", "0", "--delay", "0"]):
        pickup.run()
    mocked_transport().send.assert_has_calls(
        [
            mock.call(
                "processing_recipe",
                {
                    "recipes": [f"thing{i}"],
                    "parameters": {"foo": i},
                },
                headers=mock.ANY,
            )
            for i in range(10)
        ]
    )



from __future__ import annotations

import json
from unittest import mock

import pytest
from workflows.recipe import Recipe
from workflows.transport.offline_transport import OfflineTransport

import zocalo.configuration
from zocalo.service.dispatcher import Dispatcher


@pytest.fixture
def mock_zocalo_configuration(tmp_path):
    mock_zc = mock.MagicMock(zocalo.configuration.Configuration)
    mock_zc.storage = {
        "zocalo.recipe_directory": tmp_path,
    }
    return mock_zc


@pytest.fixture
def mock_environment(mock_zocalo_configuration):
    return {"config": mock_zocalo_configuration}


@pytest.fixture
def offline_transport(mocker):
    transport = OfflineTransport()
    mocker.spy(transport, "send")
    return transport


@pytest.fixture
def example_recipe(tmp_path):
    recipe_path = tmp_path / "example-recipe.json"
    recipe_path.write_text(
        """
{
  "1": { "service": "dispatcher test",
         "queue": "{queue}"
       },
  "start": [
     [1, { "purpose": "test for zocalo dispatcher service loading a processing recipe from an external file" }]
  ]
}
"""
    )
    return recipe_path


def test_parsing_a_custom_recipe_and_replacing_parameters(
    mock_environment, offline_transport
):
    """Passing in a recipe to the service without external dependencies.
    The recipe should be interpreted, the 'food' placeholder replaced using
    the parameter field, and the message passed back.
    The message should then contain the recipe and a correctly set pointer."""
    recipe = {
        1: {
            "service": "foo",
            "queue": "bar",
            "food": "{food}",
        },
        "start": [(1, {"purpose": "trivial test for the recipe parsing service"})],
    }
    header = {
        "message-id": mock.sentinel,
        "subscription": mock.sentinel,
    }
    parameters = {"food": "spam"}
    service = Dispatcher(environment=mock_environment)
    service.transport = offline_transport
    service.start()
    service.process(
        None, header, message={"parameters": parameters, "custom_recipe": recipe}
    )
    expected_recipe = Recipe(recipe)
    expected_recipe.apply_parameters(parameters)
    expected_message = {
        "payload": recipe["start"][0][1],
        "recipe": expected_recipe.recipe,
        "recipe-path": [],
        "recipe-pointer": 1,
        "environment": mock.ANY,
    }
    offline_transport.send.assert_called_with(
        "bar",
        expected_message,
        headers={"workflows-recipe": True},
        transaction=mock.ANY,
    )


def test_loading_a_recipe_from_a_file(
    mock_environment, offline_transport, example_recipe
):
    """When a file name is passed to the service the file should be loaded and
    parsed correctly, including parameter replacement."""
    header = {
        "message-id": mock.sentinel,
        "subscription": mock.sentinel,
    }
    parameters = {"queue": "foo"}
    service = Dispatcher(environment=mock_environment)
    service.transport = offline_transport
    service.start()
    service.process(
        None,
        header,
        message={"parameters": parameters, "recipes": [example_recipe.stem]},
    )
    with example_recipe.open() as fh:
        recipe = json.load(fh)
    expected_recipe = Recipe(recipe)
    expected_recipe.apply_parameters(parameters)
    expected_message = {
        "payload": recipe["start"][0][1],
        "recipe": expected_recipe.recipe,
        "recipe-path": [],
        "recipe-pointer": 1,
        "environment": mock.ANY,
    }
    offline_transport.send.assert_called_with(
        "foo",
        expected_message,
        headers={"workflows-recipe": True},
        transaction=mock.ANY,
    )


def test_combining_recipes(mock_environment, offline_transport, example_recipe):
    """Combine a recipe from a file and a custom recipe."""
    header = {
        "message-id": mock.sentinel,
        "subscription": mock.sentinel,
    }
    parameters = {"queue": "foo"}
    custom_recipe = {
        1: {"service": "dispatcher test", "queue": "bar"},
        "start": [(1, {"purpose": "test recipe merging"})],
    }
    message = {
        "parameters": parameters,
        "recipes": [example_recipe.stem],
        "custom_recipe": custom_recipe,
    }
    service = Dispatcher(environment=mock_environment)
    service.transport = offline_transport
    service.start()
    service.process(None, header=header, message=message)
    with example_recipe.open() as fh:
        recipe = json.load(fh)
    expected_recipe = Recipe(recipe)
    expected_recipe.apply_parameters(parameters)
    common_expected_message = {
        "recipe": mock.ANY,
        "recipe-path": [],
        "environment": mock.ANY,
    }
    offline_transport.send.assert_has_calls(
        [
            mock.call(
                "bar",
                {
                    **common_expected_message,
                    "recipe-pointer": 1,
                    "payload": custom_recipe["start"][0][1],
                },
                headers={"workflows-recipe": True},
                transaction=mock.ANY,
            ),
            mock.call(
                "foo",
                {
                    **common_expected_message,
                    "recipe-pointer": 2,
                    "payload": recipe["start"][0][1],
                },
                headers={"workflows-recipe": True},
                transaction=mock.ANY,
            ),
        ],
        any_order=True,
    )



from __future__ import annotations

import pytest

import zocalo.configuration
from zocalo.util import slurm


@pytest.fixture
def zocalo_configuration(mocker):
    zc = mocker.MagicMock(zocalo.configuration.Configuration)
    zc.slurm = {
        "url": "http://slurm.example.com:1234",
        "user": "foo",
        "user_token": "sometoken",
        "api_version": "v0.0.36",
    }
    return zc


@pytest.fixture
def slurm_api(zocalo_configuration):
    api = slurm.SlurmRestApi.from_zocalo_configuration(zocalo_configuration)
    return api


@pytest.fixture
def jobs_response():
    return {
        "errors": [],
        "jobs": [
            {
                "batch_host": "cs04r-sc-com13-01",
                "flags": ["JOB_WAS_RUNNING"],
                "cluster": "cluster",
                "command": "bash",
                "eligible_time": "1675327225",
                "end_time": "1706863225",
                "exit_code": 0,
                "group_id": "12345",
                "job_id": "129",
                "job_resources": {
                    "nodes": "cs04r-sc-com13-01",
                    "allocated_cpus": 1,
                    "allocated_hosts": 1,
                    "allocated_nodes": {
                        "0": {
                            "memory": 0,
                            "cpus": 1,
                            "sockets": {"0": "unassigned"},
                            "cores": {"0": "unassigned"},
                        }
                    },
                },
                "job_state": "RUNNING",
                "last_sched_evaluation": "1675327225",
                "max_cpus": "0",
                "max_nodes": "0",
                "name": "bash",
                "nodes": "cs04r-sc-com13-01",
                "tasks_per_board": "0",
                "cpus": "1",
                "node_count": "1",
                "tasks": "1",
                "partition": "wilson",
                "minimum_cpus_per_node": "1",
                "minimum_tmp_disk_per_node": "0",
                "priority": "4294901730",
                "qos": "normal",
                "start_time": "1675327225",
                "submit_time": "1675327225",
                "user_id": "123456",
                "user_name": "foo",
                "current_working_directory": "/home/foo",
            }
        ],
    }


def test_get_slurm_api_from_zocalo_configuration(slurm_api):
    assert slurm_api.url == "http://slurm.example.com:1234"
    assert slurm_api.version == "v0.0.36"
    assert slurm_api.user_name == "foo"
    assert slurm_api.user_token == "sometoken"


def test_get_slurm_api_user_token_external_file(tmp_path):
    user_token_file = tmp_path / "slurm-user-token"
    user_token_file.write_text("foobar")
    api = slurm.SlurmRestApi(
        url="http://slurm.example.com:1234",
        version="v0.0.36",
        user_name="foo",
        user_token=user_token_file,
    )
    assert api.user_token == "foobar"


def test_get_jobs(requests_mock, slurm_api, jobs_response):
    requests_mock.get(
        "/slurm/v0.0.36/jobs",
        json=jobs_response,
    )
    assert slurm_api.get_jobs() == slurm.models.JobsResponse(**jobs_response)


def test_get_job_info(requests_mock, slurm_api, jobs_response):
    requests_mock.get(
        "/slurm/v0.0.36/job/129",
        json=jobs_response,
    )
    assert slurm_api.get_job_info(129) == slurm.models.JobsResponse(**jobs_response)
